package ctbrec.sites.cam4;

import com.iheartradio.m3u8.*;
import com.iheartradio.m3u8.data.MasterPlaylist;
import com.iheartradio.m3u8.data.Playlist;
import com.iheartradio.m3u8.data.PlaylistData;
import com.iheartradio.m3u8.data.StreamInfo;
import ctbrec.AbstractModel;
import ctbrec.Config;
import ctbrec.NotImplementedExcetion;
import ctbrec.StringUtil;
import ctbrec.io.HttpException;
import ctbrec.recorder.download.HttpHeaderFactory;
import ctbrec.recorder.download.HttpHeaderFactoryImpl;
import ctbrec.recorder.download.StreamSource;
import okhttp3.Request;
import okhttp3.Response;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static ctbrec.Model.State.*;
import static ctbrec.io.HttpClient.bodyToJsonObject;
import static ctbrec.io.HttpConstants.*;
import static java.util.regex.Pattern.DOTALL;
import static java.util.regex.Pattern.MULTILINE;

public class Cam4Model extends AbstractModel {

    private static final Logger LOG = LoggerFactory.getLogger(Cam4Model.class);
    private String playlistUrl;
    private int[] resolution = null;
    private JSONObject modelInfo;

    @Override
    public boolean isOnline(boolean ignoreCache) throws IOException, ExecutionException, InterruptedException {
        if (ignoreCache) {
            try {
                modelInfo = loadModelInfo();
                if (modelInfo.optBoolean("privateRoom")) {
                    onlineState = PRIVATE;
                }
            } catch (Exception e) {
                onlineState = OFFLINE;
            }
        }
        return onlineState == ONLINE;
    }

    private JSONObject loadModelInfo() throws IOException {
        JSONObject roomState = new Cam4WsClient(Config.getInstance(), (Cam4) getSite(), this).getRoomState();
        LOG.trace(roomState.toString(2));
        String state = roomState.optString("newShowsState");
        setOnlineStateByShowType(state);
        setDescription(roomState.optString("status"));
        return roomState;
    }

    public void setOnlineStateByShowType(String showType) {
        switch (showType) {
            case "NORMAL", "ACCEPTING", "GROUP_SHOW_SELLING_TICKETS", "GS_SELLING_TICKETS", "GS_SELLING_TICKETS_UNSUCCESSFUL", "GS_GOAL_REACHED" ->
                    onlineState = ONLINE;
            case "PRIVATE_SHOW", "INSIDE_PS" -> onlineState = PRIVATE;
            case "INSIDE_GS", "GROUP_SHOW" -> onlineState = GROUP;
            case "PAUSED" -> onlineState = AWAY;
            case "OFFLINE" -> onlineState = OFFLINE;
            default -> {
                LOG.debug("############################## Unknown show type [{} {}]", this, showType);
                onlineState = UNKNOWN;
            }
        }

    }

    @Override
    public State getOnlineState(boolean failFast) throws IOException, ExecutionException {
        if (!failFast && onlineState == UNKNOWN) {
            try {
                modelInfo = loadModelInfo();
            } catch (Exception e) {
                LOG.warn("Couldn't load model details {}", e.getMessage());
            }
        }
        return onlineState;
    }

    private String getPlaylistUrl() throws IOException {
        try {
            getPlaylistUrlFromStreamUrl();
            if (StringUtil.isNotBlank(playlistUrl)) {
                return playlistUrl;
            }
        } catch (IOException e) {
            LOG.debug("Couldn't get playlist url from stream info: {}", e.getMessage());
        }
        if (modelInfo != null && modelInfo.has("hls")) {
            String hls = modelInfo.optString("hls");
            LOG.debug("Stream hls: {}", hls);
            if (StringUtil.isNotBlank(hls) && hls.startsWith("http")) {
                playlistUrl = hls;
                return playlistUrl;
            }
        }
        if (modelInfo != null && modelInfo.has("streamUUID")) {
            String uuid = modelInfo.optString("streamUUID");
            LOG.debug("Stream UUID: {}", uuid);
            String[] parts = uuid.split("-");
            if (parts.length > 3) {
                String urlTemplate = "https://cam4-hls.xcdnpro.com/{0}/cam4-origin-live/{1}_aac/playlist.m3u8";
                playlistUrl = MessageFormat.format(urlTemplate, parts[1], uuid);
                return playlistUrl;
            }
        }
        String page = loadModelPage();
        Matcher m = Pattern.compile("hlsUrl\\s*:\\s*'(.*?)'", DOTALL | MULTILINE).matcher(page);
        if (m.find()) {
            playlistUrl = m.group(1);
            return playlistUrl;
        }
        if (StringUtil.isBlank(playlistUrl)) {
            throw new IOException("Couldn't determine playlist url");
        }
        return playlistUrl;
    }

    private void getPlaylistUrlFromStreamUrl() throws IOException {
        String url = getSite().getBaseUrl() + "/rest/v1.0/profile/" + getName() + "/streamInfo";
        LOG.trace("Getting playlist url from {}", url);
        Request req = new Request.Builder() // @formatter:off
                .url(url)
                .header(USER_AGENT, Config.getInstance().getSettings().httpUserAgent)
                .header(ACCEPT, "*/*")
                .header(ACCEPT_LANGUAGE, "*")
                .header(CACHE_CONTROL, NO_CACHE)
                .header(PRAGMA, NO_CACHE)
                .header(REFERER, getUrl())
                .build(); // @formatter:on
        try (Response response = site.getHttpClient().execute(req)) {
            if (response.isSuccessful()) {
                JSONObject json = new JSONObject(bodyToJsonObject(response));
                if (LOG.isTraceEnabled()) LOG.trace(json.toString(2));
                if (json.has("canUseCDN")) {
                    if (json.getBoolean("canUseCDN")) {
                        playlistUrl = json.optString("cdnURL");
                    } else {
                        playlistUrl = json.optString("edgeURL");
                    }
                }
            } else {
                throw new HttpException(response.code(), response.message());
            }
        }
    }

    private String loadModelPage() throws IOException {
        Request req = new Request.Builder() // @formatter:off
                .url(getUrl())
                .header(USER_AGENT, Config.getInstance().getSettings().httpUserAgent)
                .header(ACCEPT, "*/*")
                .header(ACCEPT_LANGUAGE, "*")
                .header(CACHE_CONTROL, NO_CACHE)
                .header(PRAGMA, NO_CACHE)
                .header(REFERER, getUrl())
                .build(); // @formatter:on
        try (Response response = site.getHttpClient().execute(req)) {
            if (response.isSuccessful()) {
                return response.body().string();
            } else {
                throw new HttpException(response.code(), response.message());
            }
        }
    }

    @Override
    public List<StreamSource> getStreamSources() throws IOException, ExecutionException, ParseException, PlaylistException {
        MasterPlaylist masterPlaylist = getMasterPlaylist();
        List<StreamSource> sources = new ArrayList<>();
        for (PlaylistData playlist : masterPlaylist.getPlaylists()) {
            if (playlist.hasStreamInfo()) {
                StreamSource src = new StreamSource();
                src.bandwidth = playlist.getStreamInfo().getBandwidth();
                src.height = Optional.ofNullable(playlist.getStreamInfo()).map(StreamInfo::getResolution).map(res -> res.height).orElse(0);
                if (playlist.getUri().startsWith("http")) {
                    src.mediaPlaylistUrl = playlist.getUri();
                } else {
                    String baseUrl = playlistUrl.substring(0, playlistUrl.lastIndexOf('/') + 1);
                    src.mediaPlaylistUrl = baseUrl + playlist.getUri();
                }
                LOG.trace("Media playlist {}", src.mediaPlaylistUrl);
                sources.add(src);
            }
        }
        return sources;
    }

    private MasterPlaylist getMasterPlaylist() throws IOException, ParseException, PlaylistException {
        String masterPlaylistUrl = getPlaylistUrl();
        masterPlaylistUrl = masterPlaylistUrl.replace("_sfm4s", "");
        LOG.debug("Loading master playlist [{}]", masterPlaylistUrl);
        Request.Builder builder = new Request.Builder().url(masterPlaylistUrl);
        getHttpHeaderFactory().createMasterPlaylistHeaders().forEach(builder::header);
        Request req = builder.build();

        try (Response response = site.getHttpClient().execute(req)) {
            if (response.isSuccessful()) {
                InputStream inputStream = response.body().byteStream();
                PlaylistParser parser = new PlaylistParser(inputStream, Format.EXT_M3U, Encoding.UTF_8, ParsingMode.LENIENT);
                Playlist playlist = parser.parse();
                return playlist.getMasterPlaylist();
            } else {
                throw new HttpException(response.code(), "Couldn't download HLS playlist " + masterPlaylistUrl);
            }
        }
    }

    @Override
    public void invalidateCacheEntries() {
        resolution = null;
        playlistUrl = null;
    }

    @Override
    public void receiveTip(Double tokens) throws IOException {
        throw new NotImplementedExcetion("Tipping is not implemented for Cam4, yet");
    }

    @Override
    public int[] getStreamResolution(boolean failFast) throws ExecutionException {
        if (resolution == null) {
            if (failFast) {
                return new int[2];
            }
            try {
                if (!isOnline()) {
                    return new int[2];
                }
                List<StreamSource> sources = getStreamSources();
                Collections.sort(sources);
                StreamSource best = sources.get(sources.size() - 1);
                resolution = new int[]{best.width, best.height};
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                LOG.warn("Couldn't determine stream resolution for {} - {}", getName(), e.getMessage());
                resolution = new int[2];
            } catch (ExecutionException | IOException | ParseException | PlaylistException e) {
                LOG.warn("Couldn't determine stream resolution for {} - {}", getName(), e.getMessage());
                resolution = new int[2];
            }
        }
        return resolution;
    }

    @Override
    public boolean follow() throws IOException {
        String url = site.getBaseUrl() + "/profiles/addFriendFavorite?action=addFavorite&object=" + getName() + "&_=" + System.currentTimeMillis();
        Request req = new Request.Builder()
                .url(url)
                .addHeader(X_REQUESTED_WITH, XML_HTTP_REQUEST)
                .build();
        try (Response response = site.getHttpClient().execute(req)) {
            return response.isSuccessful();
        }
    }

    @Override
    public boolean unfollow() throws IOException {
        // send unfollow request
        String username = Config.getInstance().getSettings().cam4Username;
        String url = site.getBaseUrl() + "/rest/v1.0/favorites/" + username + '/' + getName();
        Request req = new Request.Builder()
                .url(url)
                .header(X_REQUESTED_WITH, XML_HTTP_REQUEST)
                .delete()
                .build();
        try (Response resp = site.getHttpClient().execute(req)) {
            return resp.isSuccessful();
        }
    }

    public void setPlaylistUrl(String playlistUrl) {
        this.playlistUrl = playlistUrl;
    }

    @Override
    public void setUrl(String url) {
        String normalizedUrl = url.toLowerCase();
        if (normalizedUrl.endsWith("/")) {
            normalizedUrl = normalizedUrl.substring(0, normalizedUrl.length() - 1);
        }
        super.setUrl(normalizedUrl);
    }

    @Override
    public HttpHeaderFactory getHttpHeaderFactory() {
        HttpHeaderFactoryImpl fac = new HttpHeaderFactoryImpl();
        Map<String, String> headers = new HashMap<>();
        headers.put(ACCEPT, "*/*");
        headers.put(ACCEPT_LANGUAGE, Locale.ENGLISH.getLanguage());
        headers.put(CONNECTION, KEEP_ALIVE);
        if (getSite() != null) {
            headers.put(ORIGIN, getSite().getBaseUrl());
            headers.put(REFERER, getSite().getBaseUrl());
        }
        headers.put(USER_AGENT, Config.getInstance().getSettings().httpUserAgent);
        fac.setMasterPlaylistHeaders(headers);
        fac.setSegmentPlaylistHeaders(headers);
        fac.setSegmentHeaders(headers);
        return fac;
    }
}
